from __future__ import annotations
from typing import Any, Optional, TypeVar, cast, TYPE_CHECKING
from datetime import date
from jsonclasses.types import Types
from jsonclasses.field_definition import FieldStorage, FieldType
from jsonclasses.types_resolver import TypesResolver
from inflection import underscore, camelize
from .utils import (ref_field_key, ref_field_keys, ref_db_field_key,
                    ref_db_field_keys)
from .coder import Coder
from .dbconf import DBConf
if TYPE_CHECKING:
    from .pymongo_object import PymongoObject
    T = TypeVar('T', bound=PymongoObject)


class Decoder(Coder):

    def decode_list(self,
                    value: list[Any],
                    cls: type[T],
                    types: Types) -> Optional[list[Any]]:
        if value is None:
            return None
        if types.definition.field_storage == FieldStorage.FOREIGN_KEY:
            return None
        elif types.definition.field_storage == FieldStorage.LOCAL_KEY:
            return [str(item) for item in value]
        else:
            item_types = TypesResolver().resolve_types(
                types.definition.raw_item_types)
            return ([self.decode_item(value=item, cls=cls, types=item_types)
                    for item in value])

    def decode_dict(self,
                    value: dict[str, Any],
                    cls: type[T],
                    types: Types) -> Optional[dict[str, Any]]:
        if value is None:
            return None
        config: DBConf = cls.dbconf
        if types.definition.field_storage == FieldStorage.FOREIGN_KEY:
            return None
        if types.definition.field_storage == FieldStorage.LOCAL_KEY:
            return ({underscore(k) if config.camelize_db_keys else k: str(v)
                    for k, v in value.items()})
        else:
            item_types = TypesResolver().resolve_types(
                types.definition.raw_item_types)
            return ({underscore(k) if config.camelize_db_keys else k:
                    self.decode_item(value=v, cls=cls, types=item_types)
                    for k, v in value.items()})

    def decode_shape(self,
                     value: dict[str, Any],
                     cls: type[T],
                     types: Types) -> dict[str, Any]:
        config: DBConf = cls.dbconf
        shape_types = cast(dict[str, Any], types.definition.shape_types)
        retval = {}
        for k, item_types in shape_types.items():
            retval[k] = self.decode_item(
                value=value[(camelize(k, False)
                             if config.camelize_db_keys else k)],
                cls=cls,
                types=item_types)
        return retval

    def decode_instance(self,
                        value: dict[str, Any],
                        cls: type[T],
                        types: Types) -> Any:
        from .pymongo_object import PymongoObject
        if types.definition.field_storage == FieldStorage.FOREIGN_KEY:
            return None
        elif types.definition.field_storage == FieldStorage.LOCAL_KEY:
            return str(value)
        else:
            return self.decode_root(
                root=value,
                cls=cast(type[PymongoObject], TypesResolver().resolve_types(
                    types.definition.instance_types,
                    config=cls.definition.config
                ).definition.instance_types))

    def decode_item(self, value: Any, cls: type[T], types: Types) -> Any:
        if value is None:
            return value
        if types.definition.field_type == FieldType.DATE:
            return date.fromisoformat(value.isoformat()[:10])
        elif types.definition.field_type == FieldType.ENUM:
            return types.definition.enum_class(value)
        elif types.definition.field_type == FieldType.LIST:
            return self.decode_list(value=value, cls=cls, types=types)
        elif types.definition.field_type == FieldType.DICT:
            return self.decode_dict(value=value, cls=cls, types=types)
        elif types.definition.field_type == FieldType.SHAPE:
            return self.decode_shape(value=value, cls=cls, types=types)
        elif types.definition.field_type == FieldType.INSTANCE:
            return self.decode_instance(value=value, cls=cls, types=types)
        else:
            return value

    def decode_root(self,
                    root: dict[str, Any],
                    cls: type[T]) -> T:
        dest = cls()
        for field in cls.definition.fields:
            if self.is_id_field(field):
                setattr(dest, 'id', str(root.get('_id')))
            elif self.is_foreign_key_storage(field):
                pass
            elif self.is_local_key_reference_field(field):
                setattr(
                    dest,
                    ref_field_key(field.name),
                    self.decode_item(
                        value=root.get(ref_db_field_key(
                            field.name, cls=cls)),
                        types=field.types,
                        cls=cls))
            elif self.is_local_keys_reference_field(field):
                setattr(
                    dest,
                    ref_field_keys(field.name),
                    self.decode_item(
                        value=root.get(ref_db_field_keys(
                            field.name, cls=cls)),
                        types=field.types,
                        cls=cls))
            else:
                setattr(
                    dest,
                    field.name,
                    self.decode_item(
                        value=root.get(camelize(field.name, False) if
                                       cls.dbconf.camelize_db_keys else
                                       field.name),
                        types=field.types,
                        cls=cls))
        setattr(dest, '_is_new', False)
        setattr(dest, '_is_modified', False)
        setattr(dest, '_modified_fields', set())
        setattr(dest, '_previous_values', {})
        return dest



from __future__ import annotations
from typing import TypeVar, Any, Union, cast
from re import search
from bson.objectid import ObjectId
from pymongo.errors import DuplicateKeyError
from inflection import camelize, underscore
from pymongo.collection import Collection
from jsonclasses.field_definition import FieldStorage
from jsonclasses.exceptions import (UniqueConstraintException,
                                    ObjectNotFoundException)
from jsonclasses.types_resolver import TypesResolver
from .pymongo_object import PymongoObject
from .query import ListQuery, SingleQuery, IDQuery
from .encoder import Encoder
from .connection import Connection
from .utils import btype_from_ftype


T = TypeVar('T', bound=PymongoObject)


def find(cls: type[T], **kwargs: Any) -> ListQuery[T]:
    return ListQuery(cls=cls, filter=kwargs)


def one(cls: type[T], **kwargs: Any) -> SingleQuery[T]:
    return ListQuery(cls=cls, filter=kwargs).first


def pymongo_id(cls: type[T], id: Union[str, ObjectId]) -> IDQuery[T]:
    return IDQuery(cls=cls, id=id)


def delete_by_id(cls: type[T], id: str) -> None:
    collection = Connection.get_collection(cls)
    deletion_result = collection.delete_one({'_id': ObjectId(id)})
    if deletion_result.deleted_count < 1:
        raise ObjectNotFoundException(
            f'{cls.__name__} with id \'{id}\' is not found.')
    else:
        return None


def delete_many(cls: type[T], *args, **kwargs) -> int:
    if len(args) == 0:
        args = ({},)
    collection = Connection.get_collection(cls)
    return collection.delete_many(*args, **kwargs).deleted_count


def _database_write(self: T) -> None:
    try:
        Encoder().encode_root(self).execute()
    except DuplicateKeyError as exception:
        result = search('index: (.+?)_1', exception._message)
        assert result is not None
        db_key = result.group(1)
        pt_key = db_key
        json_key = db_key
        if self.__class__.dbconf.camelize_db_keys:
            pt_key = underscore(db_key)
            json_key = pt_key
        if self.__class__.definition.config.camelize_json_keys:
            json_key = camelize(pt_key, False)
        raise UniqueConstraintException(
                getattr(self, pt_key), json_key) from None


def _orm_delete(self: T) -> None:
    for field in self.__class__.definition.deny_fields:
        if field.definition.field_storage == FieldStorage.LOCAL_KEY:
            key = self.__class__.definition.config.key_transformer(field)
            if getattr(self, key) is not None:

            # t = TypesResolver().resolve_types(field.definition.instance_types,
            #                                   self.__class__.definition.config)
            # types = t
            # oc = cast(type[PymongoObject], types.definition.instance_types)
            # field.foreign_field.definition.foreign_key

        field.definition.foreign_key
        pass


def _orm_restore(self: T) -> None:
    pass


def pymongofy(class_: type) -> PymongoObject:
    # do not install methods for subclasses
    if hasattr(class_, '__is_pymongo__'):
        return cast(PymongoObject, class_)
    # type marks
    setattr(class_, '__is_pymongo__', True)
    # public methods
    class_.find = classmethod(find)
    class_.one = classmethod(one)
    class_.id = classmethod(pymongo_id)
    class_._orm_delete = _orm_delete
    class_._orm_restore = _orm_restore
    # protected methods
    class_._database_write = _database_write
    connection = Connection.from_class(class_)
    if class_.definition.config.abstract:
        return class_
    def callback(coll: Collection):
        info = coll.index_information()
        for field in class_.definition.fields:
            name = field.name
            if class_.dbconf.camelize_db_keys:
                name = camelize(field.name, False)
            index = field.definition.index
            unique = field.definition.unique
            required = field.definition.required
            index_name = f'{name}_1'
            ftype = field.definition.field_type
            if unique:
                if required:
                    coll.create_index(name, name=index_name, unique=True)
                else:
                    coll.create_index(
                        name, name=index_name, unique=True,
                        partialFilterExpression={
                            name: {'$type': btype_from_ftype(ftype)}
                        })
            elif index:
                if required:
                    coll.create_index(name, name=index_name)
                else:
                    coll.create_index(
                        name, name=index_name,
                        partialFilterExpression={
                            name: {'$type': btype_from_ftype(ftype)}
                        })
            else:
                if index_name in info.keys():
                    coll.drop_index(index_name)
    connection.add_connected_callback(class_.dbconf.collection_name, callback)
    return class_



"""This module contains id query."""
from __future__ import annotations
from typing import (Union, TypeVar, Generator, Optional, Any, Generic,
                    overload, cast, TYPE_CHECKING)
from bson import ObjectId
from jsonclasses.exceptions import ObjectNotFoundException
from inflection import camelize
from .decoder import Decoder
from .connection import Connection
T = TypeVar('T')


class IDQuery(Generic[T]):

    def __init__(self,
                 cls: type[T],
                 id: Union[str, ObjectId]) -> None:
        self.cls = cls
        self.id = ObjectId(id)

    def exec(self) -> T:
        collection = Connection.get_collection(self.cls)
        fetched_object = collection.find_one({'_id': self.id})
        if fetched_object is not None:
            return Decoder().decode_root(fetched_object, self.cls)
        raise ObjectNotFoundException(
            f'{self.cls.__name__}(_id={self.id}) not found.')

    def __await__(self) -> Generator[None, None, T]:
        yield
        return self.exec()

    @property
    def optional(self) -> OptionalIDQuery:
        return OptionalIDQuery(cls=self.cls, id=self.id)


class OptionalIDQuery(IDQuery):

    def exec(self) -> Optional[T]:
        try:
            return super().exec()
        except ObjectNotFoundException:
            return None


class BaseQuery(Generic[T]):

    def __init__(self, cls: type[T]) -> None:
        self.cls = cls
        self.filter: Optional[dict[str, Any]] = None
        self.sort: Optional[list[tuple[str, int]]] = None
        self.projection: Optional[Union[list[str], dict[str, Any]]] = None
        self.skipping: Optional[int] = None
        self.limiting: Optional[int] = None

    def _exec(self) -> list[T]:
        kwargs: dict[str, Any] = {}
        if self.filter is not None:
            kwargs['filter'] = self.filter
        if self.sort is not None:
            kwargs['sort'] = self.sort
        if self.projection is not None:
            kwargs['projection'] = self.projection
        if self.skipping is not None:
            kwargs['skip'] = self.skipping
        if self.limiting is not None:
            kwargs['limit'] = self.limiting
        cursor = Connection.get_collection(self.cls).find(**kwargs)
        results = [doc for doc in cursor]
        coder = Decoder()
        return list(map(lambda doc: coder.decode_root(doc, self.cls), results))

    def __await__(self) -> Generator[None, None, list[T]]:
        yield
        return self._exec()


class ListQuery(BaseQuery, Generic[T]):

    def __init__(self,
                 cls: type[T],
                 filter: Optional[dict[str, Any]] = None) -> None:
        super().__init__(cls=cls)
        if filter is not None and filter.get('id'):
            filter['_id'] = ObjectId(filter['id'])
            del filter['id']
        self.filter = self._update_cases(filter)

    def __call__(self, filter: Optional[dict[str, Any]] = None) -> ListQuery:
        self.filter = self._update_cases(filter)
        return self

    def _update_cases(self,
                      d: Optional[dict[str, Any]]) -> Optional[dict[str, Any]]:
        if not self.cls.dbconf.camelize_db_keys:
            return d
        if d is None:
            return None
        retval: dict[str, Any] = {}
        for key, value in d.items():
            retval[camelize(key, False)] = value
        return retval

    def exec(self) -> list[T]:
        return super()._exec()

    def where(self, filter: dict[str, Any]) -> ListQuery:
        self.filter = filter
        return self

    @overload
    def order(self, sort: str) -> ListQuery: ...

    @overload
    def order(self, sort: tuple[str, int]) -> ListQuery: ...

    @overload
    def order(self, sort: list[tuple[str, int]]) -> ListQuery: ...

    def order(self, sort: Any) -> ListQuery:
        if isinstance(sort, str):
            self.sort = [(sort, 1)]
        elif isinstance(sort, tuple):
            self.sort = [cast(tuple[str, int], sort)]
        elif isinstance(sort, list):
            self.sort = sort
        else:
            self.sort = None
        return self

    @overload
    def project(self, projection: list[str]) -> ListQuery: ...

    @overload
    def project(self, projection: dict[str, Any]) -> ListQuery: ...

    def project(self, projection: Any) -> ListQuery:
        self.projection = projection
        return self

    def skip(self, skipping: int) -> ListQuery:
        self.skipping = skipping
        return self

    def limit(self, limiting: int) -> ListQuery:
        self.limiting = limiting
        return self

    @property
    def first(self) -> SingleQuery[T]:
        return SingleQuery(self)


class SingleQuery(BaseQuery, Generic[T]):

    def __init__(self, query: BaseQuery) -> None:
        super().__init__(cls=query.cls)
        self.filter = query.filter
        self.sort = query.sort
        self.projection = query.projection
        self.skipping = query.skipping
        self.limiting = 1

    def exec(self) -> T:
        results: list[T] = super()._exec()
        if len(results) < 1:
            raise ObjectNotFoundException(
                f'{self.cls.__name__}(filter={self.filter}, '
                f'sort={self.sort}, projection={self.projection}, '
                f'skipping={self.skipping}) not found.')
        return results[0]

    @property
    def optional(self) -> OptionalSingleQuery[T]:
        return OptionalSingleQuery(self)


class OptionalSingleQuery(SingleQuery, Generic[T]):

    def exec(self) -> Optional[T]:
        results: list[T] = super()._exec()
        if len(results) < 1:
            return None
        return results[0]
